package com.sanfeng.oss.serversigneddirectupload.demos.web;

import com.aliyun.sts20150401.models.AssumeRoleResponse;
import com.aliyun.sts20150401.models.AssumeRoleResponseBody;
import com.aliyun.tea.TeaException;
import com.aliyun.oss.common.utils.BinaryUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.codec.binary.Base64;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.GetMapping;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.*;

@Controller
public class WebController {

    // OSS basic information, replace with actual bucket name and region-id
    String bucket = "web-direct-upload";
    String region = "cn-hangzhou";

    String host = "http://" + bucket + ".oss-" + region + ".aliyuncs.com";
    // Specify the prefix for files uploaded to OSS.
    String upload_dir = "dir";
    // Specify the expiration time in seconds.
    Long expire_time = 3600L;

    /**
     * Generate an expiration time based on the specified duration in seconds.
     * @param seconds Duration in seconds.
     * @return ISO8601 formatted time string, e.g., "2014-12-01T12:00:00.000Z".
     */
    public static String generateExpiration(long seconds) {
        // Get the current timestamp in seconds
        long now = Instant.now().getEpochSecond();
        // Calculate the expiration timestamp
        long expirationTime = now + seconds;
        // Convert the timestamp to an Instant object and format it to ISO8601
        Instant instant = Instant.ofEpochSecond(expirationTime);
        // Define the time zone
        ZoneId zone = ZoneId.systemDefault();  // Use the system default time zone
        // Convert Instant to ZonedDateTime
        ZonedDateTime zonedDateTime = instant.atZone(zone);
        // Define the date-time format, e.g., 2023-12-03T13:00:00.000Z
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
        // Format the date-time
        String formattedDate = zonedDateTime.format(formatter);
        // Return the result
        return formattedDate;
    }

    // Initialize STS Client
    public static com.aliyun.sts20150401.Client createStsClient() throws Exception {
        // Code leakage may lead to the exposure of AccessKey and threaten the security of all resources under the account.
        // The following code example is for reference only.
        // It is recommended to use a more secure STS method. For more authentication access methods, please refer to:
        // https://help.aliyun.com/document_detail/378657.html.
        com.aliyun.teaopenapi.models.Config config = new com.aliyun.teaopenapi.models.Config()
                // Required, please ensure that the environment variable OSS_ACCESS_KEY_ID is set in the code running environment.
                .setAccessKeyId(System.getenv("OSS_ACCESS_KEY_ID"))
                // Required, please ensure that the environment variable OSS_ACCESS_KEY_SECRET is set in the code running environment.
                .setAccessKeySecret(System.getenv("OSS_ACCESS_KEY_SECRET"));
        // Endpoint reference: https://api.aliyun.com/product/Sts
        config.endpoint = "sts.cn-hangzhou.aliyuncs.com";
        return new com.aliyun.sts20150401.Client(config);
    }

    // Get STS temporary credentials
    public static AssumeRoleResponseBody.AssumeRoleResponseBodyCredentials getCredential() throws Exception {
        com.aliyun.sts20150401.Client client = WebController.createStsClient();
        // For specific interface parameter descriptions, please refer to:
        // https://help.aliyun.com/zh/ram/developer-reference/api-sts-2015-04-01-assumerole
        com.aliyun.sts20150401.models.AssumeRoleRequest assumeRoleRequest = new com.aliyun.sts20150401.models.AssumeRoleRequest()
                // Required, please ensure that the environment variable OSS_STS_ROLE_ARN is set in the code running environment.
                .setRoleArn(System.getenv("OSS_STS_ROLE_ARN"))
                .setRoleSessionName("yourRoleSessionName"); // Custom session name
        com.aliyun.teautil.models.RuntimeOptions runtime = new com.aliyun.teautil.models.RuntimeOptions();
        try {
            // Copy and run the code, please print the API return value yourself.
            AssumeRoleResponse response = client.assumeRoleWithOptions(assumeRoleRequest, runtime);
            // Credentials include the AccessKeyId, AccessKeySecret, and SecurityToken to be used later.
            return response.body.credentials;
        } catch (TeaException error) {
            // This is only for demonstration purposes. Please handle exceptions carefully in engineering projects.
            // Error message
            System.out.println(error.getMessage());
            // Diagnostic address
            System.out.println(error.getData().get("Recommend"));
            com.aliyun.teautil.Common.assertAsString(error.message);
        } catch (Exception _error) {
            TeaException error = new TeaException(_error.getMessage(), _error);
            // This is only for demonstration purposes. Please handle exceptions carefully in engineering projects.
            // Error message
            System.out.println(error.getMessage());
            // Diagnostic address
            System.out.println(error.getData().get("Recommend"));
            com.aliyun.teautil.Common.assertAsString(error.message);
        }
        return null;
    }

    @GetMapping("/get_post_signature_for_oss_upload")
    public ResponseEntity<Map<String, String>> getPostSignatureForOssUpload() throws Exception {
        AssumeRoleResponseBody.AssumeRoleResponseBodyCredentials sts_data = getCredential();

        String accesskeyid =  sts_data.accessKeyId;
        String accesskeysecret =  sts_data.accessKeySecret;
        String securitytoken =  sts_data.securityToken;
        System.out.println("sts accesskeyid:"+accesskeyid);
        System.out.println("sts accesskeysecret:"+accesskeysecret);
        System.out.println("sts securitytoken:"+securitytoken);

        // Get the date part of x-oss-credential, current date, format as yyyyMMdd
        LocalDate today = LocalDate.now();
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyyMMdd");
        String date = today.format(formatter);

        // Get x-oss-date
        ZonedDateTime now = ZonedDateTime.now().withZoneSameInstant(java.time.ZoneOffset.UTC);
        DateTimeFormatter formatter2 = DateTimeFormatter.ofPattern("yyyyMMdd'T'HHmmss'Z'");
        String x_oss_date = now.format(formatter2);

        // Step 1: Create policy.
        String x_oss_credential = accesskeyid + "/" + date + "/" + region + "/oss/aliyun_v4_request";

        ObjectMapper mapper = new ObjectMapper();

        Map<String, Object> policy = new HashMap<>();
        policy.put("expiration", generateExpiration(expire_time));

        List<Object> conditions = new ArrayList<>();

        Map<String, String> bucketCondition = new HashMap<>();
        bucketCondition.put("bucket", bucket);
        conditions.add(bucketCondition);

        Map<String, String> securityTokenCondition = new HashMap<>();
        securityTokenCondition.put("x-oss-security-token", securitytoken);
        conditions.add(securityTokenCondition);

        Map<String, String> signatureVersionCondition = new HashMap<>();
        signatureVersionCondition.put("x-oss-signature-version", "OSS4-HMAC-SHA256");
        conditions.add(signatureVersionCondition);

        Map<String, String> credentialCondition = new HashMap<>();
        credentialCondition.put("x-oss-credential", x_oss_credential); // Replace with actual access key id
        conditions.add(credentialCondition);

        Map<String, String> dateCondition = new HashMap<>();
        dateCondition.put("x-oss-date", x_oss_date);
        conditions.add(dateCondition);

        conditions.add(Arrays.asList("content-length-range", 1, 10240000));
        conditions.add(Arrays.asList("eq", "$success_action_status", "200"));
        conditions.add(Arrays.asList("starts-with", "$key", upload_dir));

        policy.put("conditions", conditions);

        String jsonPolicy = mapper.writeValueAsString(policy);

        // Step 2: Construct the string to sign (StringToSign).
        String stringToSign = new String(Base64.encodeBase64(jsonPolicy.getBytes()));
        // System.out.println("stringToSign: " + stringToSign);

        // Step 3: Calculate SigningKey.
        byte[] dateKey = hmacsha256(("aliyun_v4" + accesskeysecret).getBytes(), date);
        byte[] dateRegionKey = hmacsha256(dateKey, region);
        byte[] dateRegionServiceKey = hmacsha256(dateRegionKey, "oss");
        byte[] signingKey = hmacsha256(dateRegionServiceKey, "aliyun_v4_request");
        // System.out.println("signingKey: " + BinaryUtil.toBase64String(signingKey));

        // Step 4: Calculate Signature.
        byte[] result = hmacsha256(signingKey, stringToSign);
        String signature = BinaryUtil.toHex(result);
        // System.out.println("signature:" + signature);

        Map<String, String> response = new HashMap<>();
        // Add data to the map, specific elements refer to the PostObject interface documentation:
        // https://help.aliyun.com/zh/oss/developer-reference/postobject
        response.put("version", "OSS4-HMAC-SHA256");
        // This is a common mistake point, do not pass policy directly, need to do Base64 encoding
        response.put("policy", stringToSign);
        response.put("credential", x_oss_credential);
        response.put("ossdate", x_oss_date);
        response.put("signature", signature);
        response.put("token", securitytoken);
        response.put("dir", upload_dir);
        response.put("host", host);
        // Return ResponseEntity with status code 200 (OK) to the Web end for PostObject operation
        return ResponseEntity.ok(response);
    }

    public static byte[] hmacsha256(byte[] key, String data) {
        try {
            // Initialize HMAC key specification, specify the algorithm as HMAC-SHA256 and use the provided key.
            SecretKeySpec secretKeySpec = new SecretKeySpec(key, "HmacSHA256");

            // Get Mac instance, specify using HMAC-SHA256 algorithm through getInstance method.
            Mac mac = Mac.getInstance("HmacSHA256");
            // Initialize Mac object with the key.
            mac.init(secretKeySpec);

            // Execute HMAC calculation, receive the data to be calculated through doFinal method and return the result array.
            byte[] hmacBytes = mac.doFinal(data.getBytes());

            return hmacBytes;
        } catch (Exception e) {
            throw new RuntimeException("Failed to calculate HMAC-SHA256", e);
        }
    }
}
